from unittest.mock import patch
from utz import cd, dirname, join, line, match
from utz.version import VERSION_TAG_REGEX, pkg_version


def test_setup_gsmo():
    cur_dir = dirname(__file__)
    gsmo_dir = join(cur_dir,'data','gsmo')
    with cd(gsmo_dir):
        from utz.setup import setup
        import setuptools
        with patch.object(setuptools, 'setup', return_value=None) as mock_setup:
            setup()

        mock_setup.assert_called_once_with(
            **{
                'name': 'gsmo',
                'version': '0.1.1rc2',
                'description': 'Commonsense Jupyter/Docker/Git integrations.',
                'long_description': '# gsmo\nCommonsense Jupyter/Docker/Git integrations.\n\n`gsmo` streamlines mounting and working with Git repositories in Docker containers.\n\nA purpose-built Docker image+container are created (see [`gsmo.yml`] for configuration options), and Git configs are embedded so that commits inside the Docker container exist outside it (with correct permissions, authorship info, etc.).\n\nLocal notebooks or scripts can be executed [non-interactively](#non-interactive) (with results automatically committed to Git), or a [Jupyter server](#jupyter-server) or [Bash shell](#bash-shell) can be booted for interactive work.\n\n## Usage\n\n### `gsmo run`: execute notebooks non-interactively <a id="non-interactive"></a>\n`gsmo` helps run notebooks and scripts in a reproducible fashion (inside Docker containers), and pass-through changes (especially Git commits) to the host machine:\n\nRunning:\n```bash\ngsmo run\n```\nin a project directory will:\n- load configs ([`gsmo.yml`])\n- build a Docker image\n- run a container from that image\n- run the project\'s `run.ipynb` notebook inside that container\n- Git-commit results  \n\n### Interactive <a id="interactive"></a>\n\n#### Jupyter Server <a id="jupyter-server"></a>\nBuild a Docker image from the current directory, and launch a Jupyter server with the current directory mounted (and various Git- and OS-level configs set, so that changes/commits are reflected on the host machine):\n```bash\ngsmo jupyter  # or: gsmo j\n```\n- runs at a "random" but stable port (derived from a hash of the module name)\n- easily configure Python/Linux environment using [`gsmo.yml`]\n\n#### Bash shell <a id="bash-shell"></a>\nBuild a Docker image from the current directory, and launch a Bash shell with it mounted (and various Git- and OS-level configs set, so that changes/commits are reflected on the host machine):\n```bash\ngsmo sh\n```\n\n## Module configuration: \n\n### `gsmo.yml` <a id="gsmo-yml"></a>\nWhen you run `gsmo` in a directory, it will look for a `gsmo.yml` file in the current directory with any of the following fields and build a corresponding Docker image:\n\n#### Docker configs\n- `name` (`str`; default: project directory\'s basename): module name; also used as repository for built Docker image\n- `pip` (`str` or `List[str]`): `pip` libraries to install\n- `apt` (`str` or `List[str]`): `apt` libraries to install\n- `env` (`str` or `List[str]`): environment variables to set\n- `env_file` (`str`): file with environment variables\n- `group` (`str` or `List[str]`): OS groups to add to the user inside the container\n  - paths are accepted, in which case the group that owns that path will be used\n- `tag` (`str` or `List[str]`): additional Docker tags within `name` repository\n- `mount` (`str` or `List[str]`): Docker mounts, in several convenient formats:\n  - `<path>`: equivalent to `<path>:/<path>`; easily pass local project subdirectories into Docker container, e.g. `home/.bashrc`, `etc/pip.conf`, etc.\n  - standard Docker `<src>:<dst>` syntax is also supported\n  - in all cases, `~` and env vars are expanded \n- `image` (`str`; default: `runsascoded/gsmo:<gsmo version>`): base Docker image to build from; `<gsmo version>` will be the pip version of `gsmo` that was installed\n- `root` (`bool`; default `False`)\n  - when set, run as `root` inside container\n  - by default, host-machine uid+gid are used\n- `dst` (`str`: default `/src`): path inside container to mount current directory to  \n\n#### `gsmo run` configs\nThese configs are passed into the Docker container / pertain to the running of a script or notebook inside the container (see [non-interactive mode](#non-interactive)):\n- `run` (`str`; default: `run.ipynb`): notebook to run\n- `dir` (`str`; default: current directory): resolve paths (incl. mounts) relative to this directory\n- `yaml` (`str` or `List[str]`): YAML string(s) with configuration settings for the module being run\n- `yaml_path` (`str` or `List[str]`): YAML file(s) with configuration settings for the module being run\n- `commit` (`str` or `List[str]`; default: `out` config dir): paths to Git commit after a run (in non-interactive mode)\n- `out` (`str`; default `nbs`): directory to write executed notebooks to\n\n#### `gsmo jupyter` configs\n\n### `Dockerfile`\nWhen building the Docker image (in any of the above modes), if a `Dockerfile` is present in the repository, it will be built and used as the base image (and any `gsmo.yml` configs applied on top of it).\n\n[`gsmo.yml`]: #gsmo-yml\n',
                'long_description_content_type': 'text/markdown',
                'install_requires': 'pytz\n',
                # TODO: LICENSE file wasn't actually present in this release; update this test (and corresponding test submodule) to a newer release when possible
                'license': None,
                'author': 'Ryan Williams',
                'author_email': 'ryan@runsascoded.com',
                'packages': ['gsmo', 'gsmo.util'],
                'classifiers': ['Programming Language :: Python :: 3', 'Operating System :: OS Independent'],
                'python_requires': '>=3.9',
            }
        )


def test_parse_version():
    def parse(version):
        return match(VERSION_TAG_REGEX, version).groupdict()

    assert parse('0.1.2') == { 'version': '0.1.2', 'base': '0.1.2', 'commits_ahead': None, 'rc': None, 'sha': None, }
    assert parse('0.1.23r1') == { 'version': '0.1.23r1', 'base': '0.1.23', 'commits_ahead': None, 'rc': 'r1', 'sha': None, }
    assert parse('0.1.23rc45') == { 'version': '0.1.23rc45', 'base': '0.1.23', 'commits_ahead': None, 'rc': 'rc45', 'sha': None, }

    v = line('git','describe','--tags','e4282e1')
    assert v == 'v0.3.7rc1-4-ge4282e1'
    assert parse(v) == { 'version': '0.3.7rc1-4-ge4282e1', 'base': '0.3.7', 'commits_ahead': '4', 'rc': 'rc1', 'sha': 'e4282e1', }


def test_pkg_version():
    assert pkg_version('python-dateutil') == '2.9.0'



from utz.collections import *
from utz.test import raises


def test_singleton():
    assert singleton([123]) == 123
    with raises(Expected1FoundN, '2 elems found: 123,456'):
        singleton([123, 456])

    assert singleton([123, 123], dedupe=True) == 123
    with raises(Expected1FoundN, '2 elems found: 123,123'):
        singleton([123, 123], dedupe=False)

    with raises(Expected1Found0, 'No elems found'):
        singleton([])

    assert singleton([["aaa"]]) == ["aaa"]
    assert solo([["aaa"]]) == ["aaa"]

    assert singleton({ 'a': 1 }) == ('a', 1)
    with raises(Expected1FoundN, r"2 elems found: (?:\('a', 1\),\('b', 2\)|\('b', 2\),\('a', 1\))"):
        singleton({ 'a': 1, 'b': 2, })
    with raises(Expected1Found0, 'No elems found'):
        singleton({})

    assert singleton({ 'aaa': 111 }) == ('aaa', 111)
    assert singleton({ 'aaa': 111 }, key='aaa') == 111
    with raises(WrongKey, 'Expected key "bbb", found "aaa"'):
        assert singleton({ 'aaa': 111 }, key='bbb') == 111

    assert solo([ { 'a': 1, 'b': 2 } ]) == { 'a': 1, 'b': 2 }
    with raises(Expected1FoundN, '3 elems found: 1,1,1'):
        assert solo([ 1, 1, 1 ]) == 1
    with raises(Expected1FoundN, '3 elems found: 1,2,1'):
        assert solo([ 1, 2, 1 ])
    with raises(Expected1FoundN, '3 elems found: 1,1,2'):
        assert solo([ 1, 1, 2 ])

    assert solo([2, 3, 4], pred=lambda n: n % 2) == 3
    assert solo([{'a': 1}, {'b': 2}], pred=lambda o: 'a' in o) == {'a': 1}



from __future__ import annotations

import re
from re import Pattern

from utz.rgx import Includes, Excludes


def check(
    pats: list[str],
    trues: list[str],
    falses: list[str],
    **kwargs,
):
    def _check(pats: list[str | Pattern]):
        incs = Includes(pats, **kwargs)
        excs = Excludes(pats, **kwargs)
        for val in trues:
            assert incs(val)
            assert not excs(val)

        for val in falses:
            assert not incs(val)
            assert excs(val)

    _check(pats)
    _check([ re.compile(pat) for pat in pats ])


def test_patterns():
    check(
        ['a.', 'b'],
        ['aa', 'bc', 'cb'],
        ['c', 'a', 'AA', 'B'],
    )


def test_ignorecase():
    check(
        ['a.', 'b'],
        ['aa', 'AA', 'B', 'bc', 'cb'],
        ['c', 'a'],
        flags=re.I,
    )


def test_fullmatch():
    check(
        ['a.', 'b'],
        ['aa', 'b'],
        ['bc', 'c', 'cb', 'a', 'AA', 'B'],
        search=False,
    )



import asyncio

from abc import ABC
from asyncio import iscoroutine
from os.path import dirname, join

import json
from typing import Iterable

import pytest
from pytest import fixture, raises

from utz import cd, env

parametrize = pytest.mark.parametrize
from subprocess import CalledProcessError
from tempfile import NamedTemporaryFile, TemporaryDirectory

from utz import proc
from utz.proc import *
from utz.process import aio


TESTS = dirname(__file__)
SCRIPT = join(TESTS, 'streams-test.sh')
STRS = ['one', 'two', 'three']


def eq(actual, expected):
    if iscoroutine(actual):
        actual = asyncio.run(actual)
    if expected is None:
        assert actual is None
    else:
        assert actual == expected


def wait(v):
    if iscoroutine(v):
        return asyncio.run(v)
    else:
        return v


class Base(ABC):
    @staticmethod
    def test_lines(mod):
        lines = mod.lines
        eq(lines('echo', '\n'.join(STRS)), STRS)
        eq(lines('echo', '-n', '\n'.join(STRS)), STRS)
        eq(lines('[', '1', '==', '2', ']', err_ok=None), None)
        eq(lines('[', '1', '==', '2', ']', err_ok=True), [])
        with raises(CalledProcessError):
            wait(lines('[', '1', '==', '2', ']'))

    @staticmethod
    def test_output(mod):
        text = mod.text
        output = mod.output
        eq(text('echo', '\n'.join(STRS)), '\n'.join(STRS + ['']))
        eq(text('echo', '-n', '\n'.join(STRS)), '\n'.join(STRS))
        eq(output('[', '1', '==', '2', ']', err_ok=None), None)
        eq(output('[', '1', '==', '2', ']', err_ok=True), b'')
        with raises(CalledProcessError):
            wait(output('[', '1', '==', '2', ']'))

    @staticmethod
    def test_expandvars(mod):
        line = mod.line
        interpolated = f'{env["USER"]} {env["HOME"]}'
        eq(line('echo $USER $HOME', shell=True), interpolated)
        eq(line('echo', '$USER $HOME'), '$USER $HOME')
        eq(line('echo', '$USER $HOME', expandvars=True), interpolated)
        with raises(ValueError, match="Can't `expand{user,vars}` in shell mode"):
            wait(line('echo $USER $HOME', shell=True, expandvars=True))

    @staticmethod
    def test_json(mod):
        obj = [{ 'a': { 'b': 123 } }]
        with NamedTemporaryFile() as f:
            path = f.name
            with open(path, 'w') as fd:
                json.dump(obj, fd)

            eq(mod.json('cat', path), obj)
            eq(mod.json('cat', 'nonexistent-file', err_ok=True), None)
            with raises(CalledProcessError):
                wait(mod.json('cat', 'nonexistent-file'))

    @staticmethod
    def test_line(mod):
        line = mod.line
        eq(line('echo', 'yay'), 'yay')
        eq(line('echo $var', env={'var': 'aaa'}), 'aaa')
        eq(line('echo yay'), 'yay')  # Single string ⟹ `shell=True`
        eq(line('echo', '-n', 'yay'), 'yay')
        eq(line('echo', ''), '')
        eq(line('echo', '-n', '', empty_ok=True), None)
        with raises(ValueError):
            wait(line('echo', '-n', ''))
        eq(line('[', '1', '==', '2', ']', err_ok=None), None)
        eq(line('[', '1', '==', '2', ']', err_ok=True), None)
        with raises(CalledProcessError):
            wait(line('[', '1', '==', '2', ']'))

    @staticmethod
    def test_check(mod):
        check = mod.check
        eq(check('which', 'echo'), True)
        eq(check('which', 'echoz'), False)

    @staticmethod
    def test_cmd_arg_flattening(mod):
        eq(
            mod.text('echo', '-n', None, STRS, ['aaa', [None, 'bbb', 'ccc']]),
            ' '.join(STRS + ['aaa', 'bbb', 'ccc', ]),
        )

    @staticmethod
    def test_interleaved_stdout_stderr(mod):
        lines = mod.lines
        eq(lines(SCRIPT), ['stdout 1', 'stdout 2'])
        eq(lines([SCRIPT]), ['stdout 1', 'stdout 2'])
        eq(lines(SCRIPT, both=True), ['stdout 1', 'stderr 1', 'stdout 2', 'stderr 2'])
        eq(lines([SCRIPT], both=True), ['stdout 1', 'stderr 1', 'stdout 2', 'stderr 2'])


class TestSubProc(Base):
    @fixture
    def mod(self):
        return proc


class TestAioProc(Base):
    @fixture
    def mod(self):
        return aio


@parametrize(
    'cmds,shell,output', [
        (['seq 10', 'head -n5'], True, '1\n2\n3\n4\n5\n'),
        ([['seq', '10'], ['head', '-n5']], False, '1\n2\n3\n4\n5\n'),
        (['seq 5'], True, '1\n2\n3\n4\n5\n'),
        ([['seq', '5']], False, '1\n2\n3\n4\n5\n'),
    ]
)
def test_pipeline(cmds, shell, output):
    assert pipeline(cmds, shell=shell) == output

    with TemporaryDirectory() as tmpdir:
        tmp_path = join(tmpdir, 'tmp.txt')
        pipeline(cmds, shell=shell, out=tmp_path)
        with open(tmp_path) as f:
            assert f.read() == output


@fixture
def tmp_text_file():
    with NamedTemporaryFile(mode='w') as f:
        f.write('\n'.join([ f'{n}' for n in range(1, 11) ]))
        f.flush()
        yield f.name


def test_pipeline_vars_no_shell(tmp_text_file):
    with env(FILE=tmp_text_file):
        assert pipeline(
            [['cat', '$FILE'], ['head', '-n5']],
            expandvars=True,
        ) == '1\n2\n3\n4\n5\n'


def test_pipeline_vars_err():
    with raises(
        CalledProcessError,
        match=r"Command 'cat \$FILE' returned non-zero exit status 1\.",
    ) as ei:
        pipeline([['cat', '$FILE']])
    assert ei.value.stderr == "cat: '$FILE': No such file or directory\n"


def test_pipeline_vars_err2(tmp_text_file):
    with (
        env(FILE=tmp_text_file),
        raises(
            CalledProcessError,
            match=r"Command 'grep 11' returned non-zero exit status 1\.",
        ) as ei,
    ):
        pipeline([f'cat {tmp_text_file}', 'grep 11'])
    assert ei.value.stderr == ""


def test_pipeline_vars_shell(tmp_text_file):
    output = pipeline(
        ['cat $FILE', 'head -n5'],
        env={ **env, 'FILE': tmp_text_file },
    )
    assert output == '1\n2\n3\n4\n5\n'


def test_interleaved_pipelines():
    assert pipeline([SCRIPT]) == 'stdout 1\nstdout 2\n'
    assert pipeline([SCRIPT, 'wc -l']) == '2\n'
    assert pipeline([SCRIPT], both=True) == 'stdout 1\nstderr 1\nstdout 2\nstderr 2\n'
    assert pipeline([SCRIPT, 'wc -l'], both=True) == '4\n'


def test_pipeline_errs_both():
    with TemporaryDirectory() as tmpdir:
        def save(name: str, lines: Iterable):
            with open(join(tmpdir, name), 'w') as f:
                print('\n'.join([ str(l) for l in lines ]), file=f)

        save('1.txt', range(10))
        save('2.txt', range(10, 0, -2))
        expected = [
            '0',
            '1',
            '\t10',
            '2',
            '3',
            '4',
            '5',
            '6',
            '7',
            '\t\t8',
            'comm: file 2 is not in sorted order',
            '\t6',
            '\t4',
            '\t2',
            '9',
            'comm: input is not in sorted order',
            '',
        ]
        with cd(tmpdir):
            assert pipeline(['comm 1.txt 2.txt'], both=True, err_ok=True).split('\n') == expected
            with raises(CalledProcessError) as ei:
                pipeline(['comm 1.txt 2.txt'], both=True)
            exc = ei.value
            assert exc.stdout.split('\n') == expected
            assert exc.stderr is None


def test_pipeline_errs():
    with TemporaryDirectory() as tmpdir:
        def save(name: str, lines: Iterable):
            with open(join(tmpdir, name), 'w') as f:
                print('\n'.join([ str(l) for l in lines ]), file=f)

        save('1.txt', range(10))
        save('2.txt', range(10, 0, -2))
        expected = [
            '0',
            '1',
            '\t10',
            '2',
            '3',
            '4',
            '5',
            '6',
            '7',
            '\t\t8',
            '\t6',
            '\t4',
            '\t2',
            '9',
            '',
        ]
        with cd(tmpdir):
            assert pipeline(['comm 1.txt 2.txt'], err_ok=True).split('\n') == expected
            with raises(CalledProcessError) as ei:
                pipeline(['comm 1.txt 2.txt'])
            exc = ei.value
            assert exc.stdout.split('\n') == expected
            assert exc.stderr.split('\n') == [
                'comm: file 2 is not in sorted order',
                'comm: input is not in sorted order',
                '',
            ]
